#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['Dispatcher', 'DispatcherServer']

import Pyro4.core
import Pyro4.naming
from Pyro4.errors import NamingError
from queue import Queue
from pyutilib.pyro.util import *

#class Dispatcher(Pyro.core.ObjBase):
class Dispatcher():

    def __init__(self, **kwds):
        #Pyro.core.ObjBase.__init__(self)
        self.default_task_queue = Queue()
        self.default_result_queue = Queue()
        self.task_queue = {}
        self.result_queue = {}
        self.verbose = kwds.get("verbose", False)
        if self.verbose is True:
           print("Verbose output enabled...")

    def shutdown(self):
        print("Dispatcher received request to shut down - initiating...")
        self.getDaemon().shutdown()

    def add_task(self, task, type=None):
        if self.verbose is True:
           print(("Received request to add task="+str(task)+"; queue type="+str(type)))
        if type is None:
            self.default_task_queue.put(task)
        else:
            if not type in self.task_queue:
                self.task_queue[type] = Queue()
            self.task_queue[type].put(task)

    def get_task(self, type=None, timeout=5):
        if self.verbose is True:
           print(("Received request to get a task from queue type="+str(type)+"; timeout="+str(timeout)+" seconds"))
        if type is None:
            return self.default_task_queue.get(block=True, timeout=timeout)
        else:
            try:
                return self.task_queue[type].get(block=True, timeout=timeout)
            except KeyError:
                return None

    def add_result(self, data, type=None):
        if self.verbose is True:
           print(("Received request to add result with data="+str(data)+"; queue type="+str(type)))
        if type is None:
            self.default_result_queue.put(data)
        else:
            if not type in self.result_queue:
                self.result_queue[type] = Queue()
            self.result_queue[type].put(data)

    def get_result(self, type=None, timeout=5):
        if self.verbose is True:
           print(("Received request to get a result from queue type="+str(type)+"; timeout="+str(timeout)))
        if type is None:
            return self.default_result_queue.get(block=True, timeout=timeout)
        else:
            try:
                return self.result_queue[type].get(block=True, timeout=timeout)
            except KeyError:
                return None

    def num_tasks(self, type=None):
        if self.verbose is True:
           print(("Received request for number of tasks in queue with type="+str(type)))
        if type is None:
            return self.default_task_queue.qsize()
        elif type in self.task_queue:
            return self.task_queue[type].qsize()
        return 0

    def num_results(self, type=None):
        if self.verbose is True:
           print(("Received request for number of results in queue with type="+str(type)))
        if type is None:
            return self.default_result_queue.qsize()
        elif type in self.result_queue:
            return self.result_queue[type].qsize()
        return 0

    def queues_with_results(self):
        if self.verbose is True:
           print("Received request for the set of queues with results")
        result = set()
        if self.default_result_queue.qsize() > 0:
            result.add(None)

        # IMPORTANT: Make sure to make a copy (via the items() call) of the queue 
        #            contents - it can change mid-iteration, due to the introduction
        #            of new named queues.
        #
        # *python3 fix: items() no longer copies, so was changed to list(items())
        #

        for queue_name, result_queue in list(self.result_queue.items()):
           if result_queue.qsize() > 0:
               result.add(queue_name)
        return result

    def get_results_all_queues(self):

        if self.verbose is True:
           print("Received request to obtain all available results from all queues")

        result = []
        
        while self.default_result_queue.qsize() > 0:
            result.append(self.default_result_queue.get(block=False, timeout=0))

        # IMPORTANT: Make sure to make a copy (via the items() call) of the queue 
        #            contents - it can change mid-iteration, due to the introduction
        #            of new named queues.
        #
        # *python3 fix: items() no longer copies, so was changed to list(items())
        #
            
        for queue_name, result_queue in list(self.result_queue.items()):

            while result_queue.qsize() > 0:
                result.append(result_queue.get(block=False, timeout=0))

        return result       

def DispatcherServer(group=":PyUtilibServer", host=None, verbose=False):
    #
    # main program
    #
    ns = get_nameserver(host)
    if ns is None:
        return

    daemon=Pyro4.core.Daemon()
    daemon.useNameServer(ns)

    try:
        ns.createGroup(group)
    except NamingError:
        pass
    try:
        ns.unregister(group+".dispatcher")
    except NamingError:
        pass

    # the default value of 200 on clusters wasn't working well for large jobs.
    Pyro4.config.PYRO_MAXCONNECTIONS = 1000 

    uri=daemon.connect(Dispatcher(verbose=verbose),group+".dispatcher")

    print("Dispatcher is ready.")
    daemon.requestLoop()
